from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

from hostingde.exceptions import ContextConditionException
from hostingde.model import Model


class DomainContactType(Enum):
    person = 'person'
    org = 'org'
    role = 'role'


@dataclass
class DomainContact(Model):
    id: Optional[str]
    handle: Optional[str]
    type: DomainContactType
    name: str
    organization: Optional[str]
    street: List[str]
    postal_code: Optional[str]
    city: str
    state: str
    country: str
    email_address: str
    phone_number: str
    fax_number: Optional[str]
    sip_uri: Optional[str]
    hidden: Optional[bool]
    usable_by_subaccount: Optional[bool]
    add_date: Optional[str]
    last_change_date: Optional[str]

    def __init__(
        self,
        type: DomainContactType,
        name: str,
        street: List[str],
        city: str,
        country: str,
        email_address: str,
        phone_number: str,
        id: Optional[str] = None,
        handle: Optional[str] = None,
        organization: Optional[str] = None,
        postal_code: Optional[str] = None,
        fax_number: Optional[str] = None,
        sip_uri: Optional[str] = None,
        hidden: bool = False,
        usable_by_subaccount: bool = False,
        add_date: Optional[str] = None,
        last_change_date: Optional[str] = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.type = type
        self.name = name
        self.street = street
        self.city = city
        self.country = country
        self.email_address = email_address
        self.phone_number = phone_number
        self.id = id
        self.handle = handle
        self.organization = organization
        self.postal_code = postal_code
        self.fax_number = fax_number
        self.sip_uri = sip_uri
        self.hidden = hidden
        self.usable_by_subaccount = usable_by_subaccount
        self.add_date = add_date
        self.last_change_date = last_change_date

        if self.type == DomainContactType.org and not self.organization:
            raise ContextConditionException('An organization contact requires the "organization" field to be set.')



from typing import Optional, List, Union

from hostingde.domain.requests.check_availability import CheckAvailabilityRequest, CheckAvailabilityResponse
from hostingde.domain.requests.register_domain import RegisterDomainRequest
from hostingde.model.domain_contact import DomainContact
from hostingde.paginator import HostingDePaginator
from hostingde.hostingde import HostingDeCore
from hostingde.job_waiter import AsynchronousClient, JobWaiter
from hostingde.model.domain import Domain, Nameserver, DomainContactRef
from hostingde.model.filter import FilterElement
from hostingde.model.job import Job
from hostingde.model.sort import SortConfiguration


class DomainClient(HostingDeCore, AsynchronousClient):

    def __init__(self, parent: HostingDeCore):
        """
        Construct a new client

        :param parent: The parent to retrieve the session from
        """
        super().__init__(parent)

    def build_uri(self, method: str) -> str:
        """
        Utility method to easily construct URLs for this service.

        :param method: The method within this service
        :return: The URL string
        """
        return self._build_uri('domain', method)

    def jobs_find(
        self,
        limit: Optional[int] = None,
        filter: Optional[FilterElement] = None,
        sort: Optional[SortConfiguration] = None,
        page: Optional[int] = None,
        *args: list,
        **kwargs: dict
    ) -> HostingDePaginator[Job]:
        """
        Retrieves a list of Job objects from the generic filtering and sorting API.

        :param limit: The limit of objects to retrieve per call. If not set, defaults to 25.
        :param filter: A filter that is applied to the query
        :param sort: Configuration how results are sorted.
        :param page: The page to retrieve. If limit is unset, 25 items will be retrieved.
        :return: An iterator that yields ZoneConfig objects.
        """
        uri = self._build_uri('domain', 'jobsFind')

        return self._iter(uri, Job, filter, limit, sort, page)

    def check_domain_name_availability(self, domain_names: Union[str, List[str]]) -> List[CheckAvailabilityResponse]:
        uri = self.build_uri('domainStatus')

        if type(domain_names) == str:
            domain_names = [domain_names]

        response = self._request(
            uri,
            CheckAvailabilityRequest(domain_names=domain_names)
        )

        data = response.json().get('responses', [])

        return list(map(lambda x: self._instance(CheckAvailabilityResponse, x), data))

    def list_domains(
        self,
        limit: Optional[int] = None,
        filter: Optional[FilterElement] = None,
        sort: Optional[SortConfiguration] = None,
        page: Optional[int] = None,
        *args: list,
        **kwargs: dict
    ) -> HostingDePaginator[Domain]:
        """
        Retrieves a list of Domain objects from the generic filtering and sorting API.

        :param limit: The limit of objects to retrieve per call. If not set, defaults to 25.
        :param filter: A filter that is applied to the query
        :param sort: Configuration how results are sorted.
        :param page: The page to retrieve. If limit is unset, 25 items will be retrieved.
        :return: An iterator that yields Zone objects.
        """

        uri = self._build_uri('domain', 'domainsFind')

        return self._iter(uri, Domain, filter, limit, sort, page)

    def list_contacts(
        self,
        limit: Optional[int] = None,
        filter: Optional[FilterElement] = None,
        sort: Optional[SortConfiguration] = None,
        page: Optional[int] = None,
        *args: list,
        **kwargs: dict
    ) -> HostingDePaginator[DomainContact]:
        """
        Retrieves a list of Domain objects from the generic filtering and sorting API.

        :param limit: The limit of objects to retrieve per call. If not set, defaults to 25.
        :param filter: A filter that is applied to the query
        :param sort: Configuration how results are sorted.
        :param page: The page to retrieve. If limit is unset, 25 items will be retrieved.
        :return: An iterator that yields Zone objects.
        """

        uri = self._build_uri('domain', 'contactsFind')

        return self._iter(uri, DomainContact, filter, limit, sort, page)

    def register_domain(
        self,
        name: str,
        contacts: List[DomainContactRef],
        nameservers: List[Nameserver],
        transfer_lock_enabled=True,
        asynchronous: Optional[bool] = None,
    ) -> Domain:
        """
        In order to create a domain, you need to send a domainCreate request. This request takes one parameter domain
        which contains all required information of a domain object.

        :param name:
        :param contacts:
        :param nameservers:
        :param transfer_lock_enabled:
        :param asynchronous:
        :return:
        """
        url = self.build_uri('domainCreate')

        response = self._request(
            url,
            RegisterDomainRequest(
                domain=Domain(
                    name=name,
                    contacts=contacts,
                    nameservers=nameservers,
                    transfer_lock_enabled=transfer_lock_enabled
                )
            ),
        )

        domain: Domain = self._instance(Domain, response.json().get('response', {}))

        if not asynchronous and domain.id is not None:
            JobWaiter(self, domain.id, 'domainCreate').wait()

        return domain


from typing import Any, Optional, Type, TypeVar

import marshmallow_dataclass
from marshmallow import post_dump, post_load, Schema
from marshmallow.fields import Field

import hostingde


def camelcase(s: str) -> str:
    parts = iter(s.split("_"))
    return next(parts) + "".join(i.title() for i in parts)


class CamelCaseSchema(Schema):
    """Schema that uses camel-case for its external representation
    and snake-case for its internal representation.
    """

    SKIP_VALUES = {None}

    def __init__(self, **kwargs: Any):
        super().__init__(unknown='EXCLUDE', **kwargs)

    def on_bind_field(self, field_name: str, field_obj: Field) -> None:
        """Use lower-camel-casing as external representation"""
        field_obj.data_key = camelcase(field_obj.data_key or field_name)

    @post_load
    def post_load(self, data: Any, **kwargs: Any) -> Any:
        data["context"] = self.context.get('client', None)
        return data

    @post_dump
    def remove_skip_values(self, data: dict, **kwargs: dict) -> dict:
        """
        Remove None values from the serialization result
        :param data: The input data (dict). Already serialized, but may contain None values
        :param kwargs: Required by overwrite
        :return: Transformed data with None values removed
        """
        return {
            key: value
            for key, value in data.items()
            if isinstance(value, list) or isinstance(value, dict) or value not in self.SKIP_VALUES
        }


T = TypeVar('T', bound='Model')


class Model:
    """Represents a basic API Resource model"""

    @classmethod
    def get_class_instance(cls: Type[T]) -> Type[T]:
        return cls

    def to_json(self) -> dict:
        return marshmallow_dataclass.class_schema(self.get_class_instance(), base_schema=CamelCaseSchema)().dump(self)

    def __init__(self, **kwargs: Any):
        """
        Default constructor that sets all kwargs as internal properties

        All models allow for the context to be set.

        :param kwargs: dictionary of parameters that this class must support
        """
        self.__dict__ = kwargs
        self.client: Optional[hostingde.HostingDeClient] = None

        # Context might be provided as a kwarg
        if "context" in kwargs and isinstance(kwargs["context"], hostingde.HostingDeClient):
            self.client = kwargs["context"]

    @classmethod
    def from_json(cls: Type[T], data: dict, client: 'hostingde.HostingDeClient' = None) -> T:
        """
        Create a new object from a given JSON.
        :param data: The json data
        :param client: The client to use for this object
        :return: New instance of the class
        """
        return marshmallow_dataclass.class_schema(cls, base_schema=CamelCaseSchema)(context={'client': client}).load(
            data
        )



from dataclasses import dataclass, field
from typing import List, Optional

from hostingde.model import Model
from hostingde.model.record import Record


def record_list() -> List[Record]:
    return []


@dataclass
class UpdateRecordsRequest(Model):

    zone_config_id: Optional[str]
    zone_config_name: Optional[str]
    records_to_add: Optional[List[Record]] = field(default_factory=record_list)
    records_to_modify: Optional[List[Record]] = field(default_factory=record_list)
    records_to_delete: Optional[List[Record]] = field(default_factory=record_list)



from dataclasses import dataclass
from typing import Any, Optional

from hostingde.model import Model


@dataclass
class DomainSettings(Model):
    default_contact_admin_id: str
    default_contact_owner_id: str
    default_contact_tech_id: str
    default_contact_zone_id: str

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass
class Account(Model):
    """
    Name
    Name of the Account

    id
    ID of this Account.
    """

    name: Optional[str]
    id: Optional[str]
    domain_settings: Optional[DomainSettings]

    def __init__(
            self,
            name: Optional[str],
            id: Optional[str],
            domain_settings: Optional[DomainSettings] = None,
            **kwargs: Any
    ):
        """
        The Account object a object.
        :param name: Namw of this Account

        :param id: ID of this Account.
        """
        super().__init__(**kwargs)
        self.id: Optional[str] = id
        self.name: Optional[str] = name
        self.domain_settings = domain_settings



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, List, Any

from marshmallow_enum import EnumField

from hostingde.model import Model


class DomainStatus(Enum):
    ordered = 'ordered'
    active = 'active'
    restorable = 'restorable'
    failed = 'failed'
    restricted = 'restricted'


class DomainContactRefType(Enum):
    owner = 'owner'
    tech = 'tech'
    admin = 'admin'
    zone = 'zone'


@dataclass
class DomainContactRef(Model):
    contact: str
    type: DomainContactRefType = EnumField(DomainContactRefType)

    def __init__(self, contact: str, type: DomainContactRefType, **kwargs):
        super().__init__(**kwargs)

        self.contact = contact
        self.type = type


@dataclass
class Nameserver(Model):
    def __init__(self, name: str, ips: Optional[List[str]] = None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.ips = ips

    name: str
    ips: List[str] = field(default_factory=list)


@dataclass
class Domain(Model):

    id: Optional[str]
    name: str
    name_unicode: Optional[str]
    account_id: Optional[str]
    transfer_lock_enabled:  bool
    auth_info: Optional[str]
    create_date: Optional[str]
    current_contract_period_end: Optional[str]
    next_contract_period_start: Optional[str]
    deletion_type: Optional[str]
    deletion_date: Optional[str]
    add_date: Optional[str]
    last_change_date: Optional[str]
    status: Optional[DomainStatus] = EnumField(DomainStatus)
    contacts: List[DomainContactRef] = field(default_factory=list)
    nameservers: List[Nameserver] = field(default_factory=list)

    def __init__(
            self,
            name: str,
            transfer_lock_enabled: bool = True,
            id: Optional[str] = None,
            name_unicode: Optional[str] = None,
            account_id: Optional[str] = None,
            status: Optional[DomainStatus] = None,
            auth_info: Optional[str] = None,
            contacts: Optional[List[DomainContactRef]] = None,
            nameservers: Optional[List[Nameserver]] = None,
            create_date: Optional[str] = None,
            current_contract_period_end: Optional[str] = None,
            next_contract_period_start: Optional[str] = None,
            deletion_type: Optional[str] = None,
            deletion_date: Optional[str] = None,
            add_date: Optional[str] = None,
            last_change_date: Optional[str] = None,
            **kwargs: Any
    ):
        """
        Represents a domain object
        :param name: Domain name. (Unicode or ASCII format). The name will always be returned in ASCII/ACE format.
        :param transfer_lock_enabled: 	If set to true, outgoing transfers will be blocked. Please keep in mind that
                                        not every TLD supports this feature.
        :param id: Domain ID
        :param name_unicode: Domain name in Unicode/international format
        :param account_id: The id of the account
        :param status: Domain status, 'ordered', 'active', 'restorable', 'failed'
        :param auth_info: 	The domain transfer authorisation code for this domain.
                            Only contains a value if transferLockEnabled is not set.
        :param contacts: 	List of contacts for this domain
        :param nameservers: List of name servers for this domain
        :param create_date: Date of domain creation. If the domain is transferred in and the registry does not support
                            getting the original creation date, it will be the date of the transfer.
        :param current_contract_period_end:  End date of current contract period
        :param next_contract_period_start: 	Start date of next contract period
        :param deletion_type: Removal mode (delete or withdraw). Empty when domain is not scheduled for removal.
        :param deletion_date: 	Date the domain is scheduled for deletion or withdrawal. Is empty if domain is not
                                scheduled for removal.
        :param add_date: Date and time this domain was created in or transferred into our system.
        :param last_change_date: Date and time of last modification of any domain data in our system.
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.name = name
        self.transfer_lock_enabled = transfer_lock_enabled
        self.id = id
        self.name_unicode = name_unicode
        self.account_id = account_id
        self.status = status
        self.auth_info = auth_info
        self.contacts = contacts or []
        self.nameservers = nameservers or []
        self.create_date = create_date
        self.current_contract_period_end = current_contract_period_end
        self.next_contract_period_start = next_contract_period_start
        self.deletion_type = deletion_type
        self.deletion_date = deletion_date
        self.add_date = add_date
        self.last_change_date = last_change_date





from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from marshmallow_enum import EnumField

from hostingde.model import Model


class CheckAvailabilityStatus(Enum):
    alreadyRegistered = 'alreadyRegistered'
    registered = 'registered'
    nameContainsForbiddenCharacter = 'nameContainsForbiddenCharacter'
    available = 'available'
    suffixDoesNotExist = 'suffixDoesNotExist'
    suffixCannotBeRegistered = 'suffixCannotBeRegistered'
    canNotCheck = 'canNotCheck'
    unknown = 'unknown'


@dataclass
class CheckAvailabilityRequest(Model):
    domain_names: List[str]


@dataclass
class CheckAvailabilityResponse(Model):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    domain_name: str
    domain_name_unicode: str
    domain_suffix: str
    transfer_method: Optional[str]
    status: CheckAvailabilityStatus = EnumField(CheckAvailabilityStatus)



import time
from abc import ABC, abstractmethod
from typing import Optional

from hostingde.model.filter import FilterCondition, FilterElement
from hostingde.model.job import Job
from hostingde.model.sort import SortConfiguration
from hostingde.paginator import HostingDePaginator


class AsynchronousClient(ABC):
    @abstractmethod
    def jobs_find(
        self,
        limit: Optional[int] = None,
        filter: Optional[FilterElement] = None,
        sort: Optional[SortConfiguration] = None,
        *args: list,
        **kwargs: dict
    ) -> HostingDePaginator[Job]:
        """
        Retrieves a list of Job objects from the generic filtering and sorting API.

        :param limit: The limit of objects to retrieve per call. If not set, defaults to 25.
        :param filter: A filter that is applied to the query
        :param sort: Configuration how results are sorted.
        :return: An iterator that yields ZoneConfig objects.
        """
        pass


class JobWaiter:
    def __init__(self, service: AsynchronousClient, id: str, action_name: Optional[str] = None):
        self.service = service
        self.id = id
        self.action_name = action_name

    def wait(self) -> None:
        f = FilterCondition('jobObjectId').eq(self.id) \
            & FilterCondition('jobStatus').ne('successful') \
            & FilterCondition('jobStatus').ne('failed') \
            & FilterCondition('jobStatus').ne('canceled')

        # Append an additional action name filter, if provided
        if self.action_name:
            f = f & FilterCondition('jobType').eq(self.action_name)

        while True:
            jobs = self.service.jobs_find(
                filter=f
            ).fetchall()

            if len(jobs) == 0:
                break

            time.sleep(1)
